module hrt_builder
 
contains
  !     -------------------------------------------------------------------
  !!     Minimize c'x +gamma^2x'x+p'p
  !!     subject Ax+delta.p=b, x.ge.0
  !!     (following Chen et all.)
  subroutine PD_lp_bm       (nm,nd,a,na,c,x,b,irow,icol,iterlp,itercg,tol,delta,gamma,mu,PI,DI,DG,buildm,&
                             linear,m,nv,np,nt)
    use sep
    implicit none 
    integer                ::nm,nd,  na,                iterlp,itercg
    integer,dimension(:)  ::       a,         irow,icol
    real   ,dimension(:)  ::           c,x,b
    real            ::                                                tol,gamma,mu,PI,DI,DG

    real,allocatable,dimension(:):: D,v,t,r,z,y,rhs,xp,yp,dx,dy,dz
    real                         :: delta,rhod,rhop,x_one,z_one,PI_t,DI_t,DG_t
    integer :: itmax,k,i,kz,kx
    real    :: step_p,step_d
    real,dimension(:),optional:: m
    integer,optional:: np,nv,nt
    integer::buildm
    logical,optional::linear

    itmax=iterlp
    PI_t=PI       ! _t will be the comparisson standards
    DI_t=DI 
    DG_t=DG

    !! Set inital variables (D is not the data- it's a big diagonal matrix
    allocate( t(nm), v(nm),D(nm),z(nm), xp(nm)); t=0.; v=0.;d=0.;z=1.; xp=0.
    allocate(yp(nd),dy(nd),y(nd),r(nd),rhs(nd));yp=0.;dy=0.;y=1.;r=0.;rhs=0.

    iterlp=0
    do k=1,itmax
       write(0,*) ' '
       write(0,*) ' -------------------------------------------'
       write(0,*) '    Primal-Dual iteration',k
       iterlp=iterlp+1
       !!  Form least squares system to solve
       !! Develop LS equation for Interior point search direction
       call  mat_mul (1,a,na,nm,xp,nd,y,irow,icol) !!xp=A'y
       t   =c   +gamma*gamma*x   -z   -xp
       v   =mu-x   *z
       d   =1./(z   /x   +gamma*gamma)
       call  mat_mul (0,a,na,nm,x,nd,yp,irow,icol)!!yp=Ax
       r   =b   -yp   -delta*delta*y

       !!     right side term of the system
       xp   =d   *(v   /x   -t   )
       call  mat_mul (0,a,na,nm,xp,nd,yp,irow,icol) !!yp=Axp
       rhs   =r   -yp

       !!     solve ADA'dy=rhs least squares system for dual variable update (dy)
       call CG_solver(D,a,na,nm,nd,irow,icol,dy,rhs,delta,itercg,tol)

       !!    set dx and dz from dy answer
       call  mat_mul (1,a,na,nm,xp,nd,dy,irow,icol) !!xp=A'dy
       allocate(dx(nm),dz(nm));dx=0.;dz=0.
       dx   =d   *xp   +d   *(v   /x   -t   )
       dz   =v   /x   -z   *dx   /x   

       !!    primal and dual step sizes        
       x_one=x(1)+dx(1)
       z_one=z(1)+dz(1)
       kx=1
       kz=1
       do i=2,nm
          if(z_one.ge.(z(i)+dz(i))) then
             z_one=z(i)+dz(i)
             kz=i
          endif
          if(x_one.ge.(x(i)+dx(i))) then 
             x_one=x(i)+dx(i)
             kx=i
          endif
       enddo

       if(x_one.ge.0.) step_p=1.
       if(x_one.lt.0.) step_p=-x(kx)/dx(kx)

       if(z_one.ge.0.) step_d=1.
       if(z_one.lt.0.) step_d=-z(kz)/dz(kz)

       rhop=.99*step_p
       rhod=.99*step_d

       x   =x   +rhop*dx   
       z   =z   +rhod*dz
       y   =y   +rhod*dy   


       mu=(1.- min(rhop,rhod,0.99))*mu
       write(0,*) 'mu should be approaching zero:',mu  

       !!       primal-dual (convergence) condition test 
       PI=dot(nd,r,r)/(dot(nm,x,x)+1.0)
       DI=dot(nm,t,t)/(dot(nd,y,y)+1.0)
       DG=dot(nm,z,x)/(dot(nm,x,x)+1.0)
       write(0,*) ' '
       write(0,*) 'Current Daulity Gap =',DG
       write(0,*) '             DG goal=',DG_t
       if(PI.lt.PI_t.and.DI.lt.DI_t .and. DG.lt.DG_t) then
          write(0,*) 'Converged to feasability and duality gap parameters.'
          return
       end if
       if(buildm==1) then
          do i=1,int(nm/2)
             m(i) = x(i)-x(i+int(nm/2))
          end do
          call sseek('out',0,0)
          call sep_write(m(1:nv*nt))
          if (linear) then
             call sseek('rayp',0,0)
             call sep_write(m(nv+1:nv*nt+np*nt),'rayp')
          end if
             
       end if
    end do !main iteration loop
  end subroutine PD_lp_bm



  !!        -------------------------------------

  real function dot( n, x, y)
    implicit none
    real,dimension(:) ::x, y
    integer:: n,i
    real   ::val 
    dot=-999
    val=0.0

    do i=1,n
       val = val + x(i)*y(i)
    enddo

    dot=val

    return
  end function dot
  !!     ----------------------------------------
  !!    Sparse matrix multiplication when operators
  !!    can be stored in sparse matrix format
  !!    maybe a bit more efficient than operators
  subroutine mat_mul        (conj,a,na,nx,x,ny,y,irow,icol)
    implicit none
    integer                ::conj,  na,nx,  ny            ,k
    integer,dimension(:) ::a,             irow,icol
    real   ,dimension(:)::        x,   y

    if(conj.eq.0) then
       y=0.0
       do k=1,na
          y(irow(k))=y(irow(k))+a(k)*x(icol(k))
       enddo
    else
       x=0.0
       do k=1,na
          x(icol(k))=x(icol(k))+a(k)*y(irow(k))
       enddo
    endif
  end subroutine mat_mul
  !!---------------------------------------------------------------
  !!     Solve the system
  !!                        ADA'x=b when A is sparse and D is
  !!                        diagonal      
  !!     use in each iteration of PD_lp_bm - 
  subroutine  CG_solver(D,a,na,nm,nd,irow,icol,x,b,delta,itercg,tol)
    implicit none
    integer        ::       na,nm,nd,                    itercg
    real           ::                              delta,       tol
    real ,dimension(:)::D,                     x,b
    integer,dimension(:)::a, irow,icol
    real           :: beta,rho,rho_old, alpha
    integer        :: iter_max,countCG,niter,iter
    real ,allocatable,dimension(:)  ::  r,p,w,xp

    iter_max = na
    allocate(r(nd),p(nd),w(nd) )
    allocate(xp(nm))
    x=0. !! first model
    r=b  !! first residual
    rho=dot(nd,r,r)
    countCG=0
    niter=min(itercg,iter_max)

    do iter=1,niter
       countCG=countCG+1
       if(iter.eq.1) then
          p=r
       else
          beta=rho/rho_old
          p=r+beta*p
       endif
       !!        xp=A'p
       call  mat_mul (1,a,na,nm,xp,nd,p,irow,icol)
       !!        D.xp
       xp=D*xp
       call  mat_mul (0,a,na,nm,xp,nd,w,irow,icol)
       !!       Regularization
       w=w+p*delta**2
       alpha=rho/dot(nd,p,w)
       x=x+alpha*p
       r=r-alpha*w
       rho_old=rho
       rho=dot(nd,r,r)
       if(sqrt(rho/nd).le.tol) then
          write(0,*) ' '
          write(0,*)'CG iterations converged below tolerance of',tol 
          write(0,*) countCG,'iterations required.'
          write(0,*) ' '
          deallocate(r,p,w,xp)
          return
       end if
    enddo  ! main loop
    deallocate(r,p,w,xp)
    write(0,*) ' '
    write(0,*) "CG iterations insufficient to reach tolerance",tol
    write(0,*) "              stopped at ",sqrt(rho/nd)
    write(0,*) countCG,' CG iterations used for barrier step.'
    write(0,*) ' '
  end subroutine CG_solver

  subroutine HRTmat(nh,dh,h0,nt,dt,t0,nv,dv,v0,t_icol,t_irow,t_a,na,linear,p0,dp,np)
    implicit none
    real,optional::                                                        p0,dp
    integer,optional::                                                           np
    integer      :: nh,      nt,      nv,                        na
    real         ::    dh,h0,   dt,t0,   dv,v0
    integer,dimension(:) :: t_icol,t_irow,t_a
    real                 :: t, tau, v, h, p
    integer              :: it,itau,iv,ih,ip,count,j,k
    logical::linear
    count=0
    do itau=1,nt
       tau=t0+(itau-1)*dt
       do ih=1,nh
          h=h0+(ih-1)*dh
          do iv=1,nv              !!all velo's
             v=v0+(iv-1)*dv
             t=(sqrt(tau*tau+h*h/(v*v))-t0)/dt
             it=1+int(t)
             if(it.gt.0) then     !!stop at border conditions
                if(it.lt.nt) then 
                   count=count+1
                   j =(ih-1)*nt+it
                   k =(iv-1)*nt+itau
                   t_irow(count)=j
                   t_icol(count)=k
                   t_a(count)=1
                endif
             endif
          enddo
       enddo
    enddo
    if (linear) then
       do itau=1,nt
          tau=t0+(itau-1)*dt
          do ih=1,nh
             h=h0+(ih-1)*dh
             do ip=1,np              !!all p's
                p=p0+(ip-1)*dp
                t=(tau+h*p-t0)/dt
                it=1+int(t)
                if(it.gt.0) then     !!stop at border conditions
                   if(it.lt.nt) then 
                      count=count+1
                      j =(ih-1)*nt+it
                      k =(nv+ip-1)*nt+itau !! add nv to get to columns after HRT above
                      t_irow(count)=j
                      t_icol(count)=k
                      t_a(count)=1
                   endif
                endif
             enddo
          enddo
       enddo
 end if
    na=count
  end subroutine HRTmat

end module hrt_builder
